@@ -295,26 +295,42 @@ static void device_evacuate_usage(void)
 	     "Usage: bcachefs device evacuate [OPTION]... device\n"
 	     "\n"
 	     "Options:\n"
-	     "  -f, --force                  Force if data redundancy will be degraded\n"
 	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 }
 
+static int evacuate_v0(struct bchfs_handle fs, unsigned dev_idx, const char *dev_path)
+{
+	struct bch_ioctl_dev_usage_v2 *u = bchu_dev_usage(fs, dev_idx);
+
+	if (u->state == BCH_MEMBER_STATE_rw) {
+		printf("Setting %s readonly\n", dev_path);
+		bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_ro, BCH_FORCE_IF_DEGRADED);
+	}
+
+	free(u);
+
+	return bchu_data(fs, (struct bch_ioctl_data) {
+		.op		= BCH_DATA_OP_migrate,
+		.start_btree	= 0,
+		.start_pos	= POS_MIN,
+		.end_btree	= BTREE_ID_NR,
+		.end_pos	= POS_MAX,
+		.migrate.dev	= dev_idx,
+	});
+}
+
 static int cmd_device_evacuate(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
-		{ "force",			no_argument,	NULL, 'f' },
 		{ "help",			no_argument,	NULL, 'h' },
 		{ NULL }
 	};
-	int opt, flags = 0;
+	int opt;
 
 	while ((opt = getopt_long(argc, argv, "fh", longopts, NULL)) != -1)
 		switch (opt) {
-		case 'f':
-			flags |= BCH_FORCE_IF_DEGRADED;
-			break;
 		case 'h':
 			device_evacuate_usage();
 			exit(EXIT_SUCCESS);
@@ -331,23 +347,34 @@ static int cmd_device_evacuate(int argc, char *argv[])
 	int dev_idx;
 	struct bchfs_handle fs = bchu_fs_open_by_dev(dev_path, &dev_idx);
 
-	struct bch_ioctl_dev_usage_v2 *u = bchu_dev_usage(fs, dev_idx);
+	if (bcachefs_kernel_version() < bcachefs_metadata_version_reconcile)
+		return evacuate_v0(fs, dev_idx, dev_path);
 
-	if (u->state == BCH_MEMBER_STATE_rw) {
-		printf("Setting %s readonly\n", dev_path);
-		bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_ro, flags);
+	printf("Setting %s evacuating \n", dev_path);
+	bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_evacuating, BCH_FORCE_IF_DEGRADED);
+
+	while (true) {
+		struct bch_ioctl_dev_usage_v2 *u = bchu_dev_usage(fs, dev_idx);
+
+		u64 data = 0;
+		for (unsigned type = 0; type < u->nr_data_types; type++)
+			if (!data_type_is_empty(type) &&
+			    !data_type_is_hidden(type))
+				data += u->d[type].sectors;
+		free(u);
+
+		printf("\33[2K\r");
+		CLASS(printbuf, buf)();
+		prt_units_u64(&buf, data << 9);
+
+		fputs(buf.buf, stdout);
+		fflush(stdout);
+
+		if (!data)
+			return 0;
+
+		sleep(1);
 	}
-
-	free(u);
-
-	return bchu_data(fs, (struct bch_ioctl_data) {
-		.op		= BCH_DATA_OP_migrate,
-		.start_btree	= 0,
-		.start_pos	= POS_MIN,
-		.end_btree	= BTREE_ID_NR,
-		.end_pos	= POS_MAX,
-		.migrate.dev	= dev_idx,
-	});
 }
 
 static void device_set_state_usage(void)
@@ -355,7 +382,7 @@ static void device_set_state_usage(void)
 	puts("bcachefs device set-state\n"
 	     "Usage: bcachefs device set-state <new-state> <device>|<devid> <path>\n"
 	     "\n"
-	     "<new-state>: one of rw, ro, failed or spare\n"
+	     "<new-state>: one of rw, ro, evacuating or spare\n"
 	     "<path>: path to mounted filesystem, optional unless specifying device by id\n"
 	     "\n"
 	     "Options:\n"
@@ -416,27 +443,37 @@ static int cmd_device_set_state(int argc, char *argv[])
 		by_id = true;
 
 	if (offline) {
-		struct bch_opts opts = bch2_opts_empty();
-		struct bch_sb_handle sb = { NULL };
-
 		if (by_id)
 			die("Cannot specify offline device by id");
 
+		struct bch_opts opts = bch2_opts_empty();
+		opt_set(opts, nostart,	true);
+		opt_set(opts, degraded, BCH_DEGRADED_very);
+
+		struct bch_sb_handle sb = { NULL };
 		int ret = bch2_read_super(dev_str, &opts, &sb);
 		if (ret)
 			die("error opening %s: %s", dev_str, bch2_err_str(ret));
 
-		struct bch_member *m = bch2_members_v2_get_mut(sb.sb, sb.sb->dev_idx);
-
-		SET_BCH_MEMBER_STATE(m, new_state);
-
-		le64_add_cpu(&sb.sb->seq, 1);
-
-		bch2_super_write(sb.bdev->bd_fd, sb.sb);
-		ret = fsync(sb.bdev->bd_fd);
-		if (ret)
-			fprintf(stderr, "error writing superblock: fsync error (%m)");
+		unsigned dev_idx = sb.sb->dev_idx;
 		bch2_free_super(&sb);
+
+		/* scan for all devices in fs */
+		darray_const_str devs = get_or_split_cmdline_devs(1, &dev_str);
+
+		struct bch_fs *c = bch2_fs_open(&devs, &opts);
+		ret = PTR_ERR_OR_ZERO(c);
+		if (ret)
+			die("Error opening filesystem: %s", bch2_err_str(ret));
+
+		scoped_guard(mutex, &c->sb_lock) {
+			struct bch_member *m = bch2_members_v2_get_mut(c->disk_sb.sb, dev_idx);
+
+			SET_BCH_MEMBER_STATE(m, new_state);
+
+			bch2_write_super(c);
+		}
+		bch2_fs_exit(c);
 		return ret;
 	}
 
@@ -564,7 +601,7 @@ static int cmd_device_resize(int argc, char *argv[])
 			fprintf(stderr, "resize error: %s\n%s", bch2_err_str(ret), err.buf);
 
 		enumerated_ref_put(&resize->io_ref[READ], 0);
-		bch2_fs_stop(c);
+		bch2_fs_exit(c);
 	}
 	return 0;
 }
@@ -663,7 +700,7 @@ static int cmd_device_resize_journal(int argc, char *argv[])
 			fprintf(stderr, "resize error: %s\n", bch2_err_str(ret));
 
 		enumerated_ref_put(&resize->io_ref[READ], 0);
-		bch2_fs_stop(c);
+		bch2_fs_exit(c);
 	}
 	return 0;
 }
@@ -679,7 +716,7 @@ static int device_usage(void)
             "  online                       Re-add an existing member to a filesystem\n"
             "  offline                      Take a device offline, without removing it\n"
             "  evacuate                     Migrate data off a specific device\n"
-            "  set-state                    Mark a device as failed\n"
+            "  set-state                    Change device state (rw, ro, evacuating, spare)\n"
             "  resize                       Resize filesystem on a device\n"
             "  resize-journal               Resize journal on a device\n"
             "\n"
