@@ -11,6 +11,7 @@
 #include "bcachefs.h"
 #include "btree/cache.h"
 #include "btree/iter.h"
+#include "btree/read.h"
 
 #include "data/extents.h"
 
@@ -73,12 +74,13 @@ struct dump_opts {
 	char		*out;
 	bool		force;
 	bool		sanitize;
+	bool		sanitize_filenames;
 	bool		entire_journal;
 	bool		noexcl;
 };
 
 static void sanitize_key(struct bkey_packed *k, struct bkey_format *f, void *end,
-			 bool *modified)
+			 struct dump_opts opts, bool *modified)
 {
 	struct bch_val *v = bkeyp_val(f, k);
 	unsigned len = min_t(unsigned, end - (void *) v, bkeyp_val_bytes(f, k));
@@ -98,10 +100,19 @@ static void sanitize_key(struct bkey_packed *k, struct bkey_format *f, void *end
 		*modified = true;
 		break;
 	}
+
+	case KEY_TYPE_dirent:
+		if (opts.sanitize_filenames) {
+			struct bch_dirent *d = container_of(v, struct bch_dirent, v);
+
+			memset(d->d_name, 'X', len - offsetof(struct bch_dirent, d_name));
+			*modified = true;
+		}
 	}
 }
 
-static void sanitize_journal(struct bch_fs *c, void *buf, size_t len)
+static void sanitize_journal(struct bch_fs *c, void *buf, size_t len,
+			     struct dump_opts opts)
 {
 	struct bkey_format f = BKEY_FORMAT_CURRENT;
 	void *end = buf + len;
@@ -113,6 +124,28 @@ static void sanitize_journal(struct bch_fs *c, void *buf, size_t len)
 		if (le64_to_cpu(j->magic) != jset_magic(c))
 			break;
 
+		if (bch2_csum_type_is_encryption(JSET_CSUM_TYPE(j))) {
+			if (!c->chacha20_key_set) {
+				fprintf(stderr,
+					"found encrypted journal entry on non-encrypted filesystem\n");
+				return;
+			}
+
+			if (vstruct_bytes(j) > len) {
+				fprintf(stderr,
+					"encrypted journal entry overruns bucket; skipping\n");
+				return;
+			}
+
+			int ret = bch2_encrypt(c, JSET_CSUM_TYPE(j), journal_nonce(j),
+					       j->encrypted_start,
+					       vstruct_end(j) - (void *) j->encrypted_start);
+			if (ret)
+				die("error deecrypting journal entry: %s", bch2_err_str(ret));
+
+			modified = true;
+		}
+
 		vstruct_for_each(j, i) {
 			if ((void *) i >= end)
 				break;
@@ -125,7 +158,7 @@ static void sanitize_journal(struct bch_fs *c, void *buf, size_t len)
 					break;
 				if (!k->k.u64s)
 					break;
-				sanitize_key(bkey_to_packed(k), &f, end, &modified);
+				sanitize_key(bkey_to_packed(k), &f, end, opts, &modified);
 			}
 		}
 
@@ -140,12 +173,14 @@ static void sanitize_journal(struct bch_fs *c, void *buf, size_t len)
 	}
 }
 
-static void sanitize_btree(struct bch_fs *c, void *buf, size_t len)
+static void sanitize_btree(struct bch_fs *c, void *buf, size_t len,
+			   struct dump_opts opts)
 {
 	void *end = buf + len;
 	bool first = true;
 	struct bkey_format f_current = BKEY_FORMAT_CURRENT;
 	struct bkey_format f;
+	unsigned offset = 0;
 	u64 seq;
 
 	while (len) {
@@ -174,13 +209,33 @@ static void sanitize_btree(struct bch_fs *c, void *buf, size_t len)
 			sectors = vstruct_sectors(bne, c->block_bits);
 		}
 
+		if (bch2_csum_type_is_encryption(BSET_CSUM_TYPE(i))) {
+			if (!c->chacha20_key_set) {
+				fprintf(stderr,
+					"found encrypted btree node on non-encrypted filesystem\n");
+				return;
+			}
+
+			if (vstruct_end(i) > end) {
+				fprintf(stderr,
+					"encrypted btree node entry overruns bucket; skipping\n");
+				return;
+			}
+
+			int ret = bset_encrypt(c, i, offset);
+			if (ret)
+				die("error deecrypting btree node: %s", bch2_err_str(ret));
+
+			modified = true;
+		}
+
 		vstruct_for_each(i, k) {
 			if ((void *) k >= end)
 				break;
 			if (!k->u64s)
 				break;
 
-			sanitize_key(k, bkey_packed(k) ? &f : &f_current, end, &modified);
+			sanitize_key(k, bkey_packed(k) ? &f : &f_current, end, opts, &modified);
 		}
 
 		if (modified) {
@@ -199,12 +254,15 @@ static void sanitize_btree(struct bch_fs *c, void *buf, size_t len)
 		unsigned b = min(len, sectors << 9);
 		len -= b;
 		buf += b;
+		offset += b;
 	}
 }
 
 static int dump_fs(struct bch_fs *c, struct dump_opts opts)
 {
-	if (opts.sanitize)
+	if (opts.sanitize_filenames)
+		printf("Sanitizing filenames inline data extents\n");
+	else if (opts.sanitize)
 		printf("Sanitizing inline data extents\n");
 
 	dump_devs devs = {};
@@ -279,7 +337,7 @@ static int dump_fs(struct bch_fs *c, struct dump_opts opts)
 				BUG_ON(len > bucket_bytes);
 
 				xpread(img.infd, buf, len, r->start);
-				sanitize_journal(c, buf, len);
+				sanitize_journal(c, buf, len, opts);
 				qcow2_write_buf(&img, buf, len, r->start);
 			}
 
@@ -288,7 +346,7 @@ static int dump_fs(struct bch_fs *c, struct dump_opts opts)
 				BUG_ON(len > bucket_bytes);
 
 				xpread(img.infd, buf, len, r->start);
-				sanitize_btree(c, buf, len);
+				sanitize_btree(c, buf, len, opts);
 				qcow2_write_buf(&img, buf, len, r->start);
 			}
 			free(buf);
@@ -317,13 +375,13 @@ static void dump_usage(void)
 	     "Usage: bcachefs dump [OPTION]... <devices>\n"
 	     "\n"
 	     "Options:\n"
-	     "  -o output                    Output qcow2 image(s)\n"
-	     "  -f, --force                  Force; overwrite when needed\n"
-	     "  -s, --sanitize               Zero out inline data extents\n"
-	     "      --nojournal              Don't dump entire journal, just dirty entries\n"
-	     "      --noexcl                 Open devices with O_NOEXCL (not recommended)\n"
+	     "  -o output                     Output qcow2 image(s)\n"
+	     "  -f, --force                   Force; overwrite when needed\n"
+	     "  -s, --sanitize=data,filenames Zero out inline data extents, or filenames and inline data\n"
+	     "      --nojournal               Don't dump entire journal, just dirty entries\n"
+	     "      --noexcl                  Open devices with O_NOEXCL (not recommended)\n"
 	     "  -v, --verbose\n"
-	     "  -h, --help                   Display this help and exit\n"
+	     "  -h, --help                    Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 }
@@ -332,7 +390,7 @@ int cmd_dump(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
 		{ "force",		no_argument,		NULL, 'f' },
-		{ "sanitize",		no_argument,		NULL, 's' },
+		{ "sanitize",		optional_argument,	NULL, 's' },
 		{ "nojournal",		no_argument,		NULL, 'j' },
 		{ "noexcl",		no_argument,		NULL, 'e' },
 		{ "verbose",		no_argument,		NULL, 'v' },
@@ -362,6 +420,15 @@ int cmd_dump(int argc, char *argv[])
 			break;
 		case 's':
 			opts.sanitize = true;
+
+			if (optarg) {
+				if (!strcmp(optarg, "data"))
+					;
+				else if (!strcmp(optarg, "filenames"))
+					opts.sanitize_filenames = true;
+				else
+					die("Bad sanitize option %s", optarg);
+			}
 			break;
 		case 'j':
 			opts.entire_journal = false;
