@@ -97,7 +97,7 @@ static void bcachefs_fuse_destroy(void *arg)
 {
 	struct bch_fs *c = arg;
 
-	bch2_fs_stop(c);
+	bch2_fs_exit(c);
 }
 
 static void bcachefs_fuse_lookup(fuse_req_t req, fuse_ino_t dir_ino,
@@ -119,7 +119,12 @@ static void bcachefs_fuse_lookup(fuse_req_t req, fuse_ino_t dir_ino,
 		return;
 	}
 
-	struct bch_hash_info hash_info = bch2_hash_info_init(c, &bi);
+	struct bch_hash_info hash_info;
+	ret = bch2_hash_info_init(c, &bi, &hash_info);
+	if (ret) {
+		fuse_reply_err(req, -ret);
+		return;
+	}
 
 	ret = bch2_dirent_lookup(c, dir, &hash_info, &qstr, &inum);
 	if (ret) {
@@ -378,10 +383,7 @@ static void userbio_init(struct bio *bio, struct bio_vec *bv,
 			 void *buf, size_t size)
 {
 	bio_init(bio, NULL, bv, 1, 0);
-	bio->bi_iter.bi_size	= size;
-	bv->bv_page		= buf;
-	bv->bv_len		= size;
-	bv->bv_offset		= 0;
+	bio_add_virt_nofail(bio, buf, size);
 }
 
 static int get_inode_io_opts(struct bch_fs *c, subvol_inum inum, struct bch_inode_opts *opts)
@@ -947,9 +949,9 @@ static void bcachefs_fuse_readdir(fuse_req_t req, fuse_ino_t dir_ino,
 	if (!handle_dots(&ctx, dir.inum))
 		goto reply;
 
-	struct bch_hash_info dir_hash = bch2_hash_info_init(c, &bi);
-
-	ret = bch2_readdir(c, dir, &dir_hash, &ctx.ctx);
+	struct bch_hash_info dir_hash;
+	ret =   bch2_hash_info_init(c, &bi, &dir_hash) ?:
+		bch2_readdir(c, dir, &dir_hash, &ctx.ctx);
 reply:
 	if (!ret) {
 		fuse_log(FUSE_LOG_DEBUG, "bcachefs_fuse_readdir reply %zd\n",
@@ -1292,7 +1294,7 @@ out:
 	return ret ? 1 : 0;
 
 err:
-	bch2_fs_stop(c);
+	bch2_fs_exit(c);
 	goto out;
 }
 
