@@ -21,17 +21,21 @@
 
 #include "libbcachefs.h"
 #include "crypto.h"
-#include "libbcachefs/bcachefs_format.h"
-#include "libbcachefs/btree_cache.h"
-#include "libbcachefs/buckets.h"
-#include "libbcachefs/checksum.h"
-#include "libbcachefs/disk_groups.h"
-#include "libbcachefs/journal_seq_blacklist.h"
-#include "libbcachefs/opts.h"
-#include "libbcachefs/replicas.h"
-#include "libbcachefs/super-io.h"
 #include "tools-util.h"
 
+#include "bcachefs.h"
+
+#include "alloc/buckets.h"
+#include "alloc/disk_groups.h"
+#include "alloc/replicas.h"
+#include "btree/cache.h"
+
+#include "data/checksum.h"
+
+#include "journal/seq_blacklist.h"
+
+#include "sb/io.h"
+
 #define NSEC_PER_SEC	1000000000L
 
 void bch2_sb_layout_init(struct bch_sb_layout *l,
@@ -732,7 +736,7 @@ const struct bch_option *bch2_cmdline_opt_parse(int argc, char *argv[],
 		return NULL;
 
 	char *optstr = strdup(argv[optind] + 2);
-	optarg = argv[optind + 1];
+	optarg = NULL;
 
 	char *eq = strchr(optstr, '=');
 	if (eq) {
@@ -740,10 +744,6 @@ const struct bch_option *bch2_cmdline_opt_parse(int argc, char *argv[],
 		optarg = eq + 1;
 	}
 
-	if (!optarg)
-		optarg = "1";
-
-
 	int optid = bch2_opt_lookup(optstr);
 	if (optid < 0)
 		goto noopt;
@@ -754,11 +754,11 @@ const struct bch_option *bch2_cmdline_opt_parse(int argc, char *argv[],
 
 	optind++;
 
-	if (opt->type != BCH_OPT_BOOL) {
-		if (optarg == argv[optind])
-			optind++;
-	} else {
-		optarg = NULL;
+	if (!optarg) {
+		if (opt->type != BCH_OPT_BOOL)
+			optarg = argv[optind++];
+		else
+			optarg = "1";
 	}
 
 	return opt;
@@ -865,7 +865,7 @@ struct bch_opts bch2_parse_opts(struct bch_opt_strs strs)
 void bch2_opts_usage(unsigned opt_types)
 {
 	const struct bch_option *opt;
-	unsigned i, c = 0, helpcol = 30;
+	unsigned i, c = 0, helpcol = 32;
 
 	for (opt = bch2_opt_table;
 	     opt < bch2_opt_table + bch2_opts_nr;
@@ -895,13 +895,14 @@ void bch2_opts_usage(unsigned opt_types)
 		if (opt->help) {
 			const char *l = opt->help;
 
-			if (c >= helpcol)
+			if (c > helpcol) {
 				newline(c);
+			}
 
 			while (1) {
 				const char *n = strchrnul(l, '\n');
 
-				while (c < helpcol) {
+				while (c < helpcol-1) {
 					putchar(' ');
 					c++;
 				}
