@@ -33,6 +33,7 @@ typedef unsigned int __bitwise blk_mode_t;
 #define BLK_OPEN_WRITE_IOCTL	((__force blk_mode_t)(1 << 4))
 
 #define BLK_OPEN_BUFFERED	((__force blk_mode_t)(1 << 5))
+#define BLK_OPEN_CREAT		((__force blk_mode_t)(1 << 6))
 
 struct inode {
 	unsigned long		i_ino;
@@ -56,6 +57,11 @@ struct hd_struct {
 
 struct block_device {
 	struct kobject		kobj;
+
+	struct			{
+		struct kobject	kobj;
+	}			bd_device;
+
 	dev_t			bd_dev;
 	char			name[BDEVNAME_SIZE];
 	struct inode		*bd_inode;
@@ -81,7 +87,7 @@ typedef u8 __bitwise blk_status_t;
 #define BLK_STS_NOSPC		((__force blk_status_t)3)
 #define BLK_STS_TRANSPORT	((__force blk_status_t)4)
 #define BLK_STS_TARGET		((__force blk_status_t)5)
-#define BLK_STS_NEXUS		((__force blk_status_t)6)
+#define BLK_STS_RESV_CONFLICT	((__force blk_status_t)6)
 #define BLK_STS_MEDIUM		((__force blk_status_t)7)
 #define BLK_STS_PROTECTION	((__force blk_status_t)8)
 #define BLK_STS_RESOURCE	((__force blk_status_t)9)
@@ -91,6 +97,12 @@ typedef u8 __bitwise blk_status_t;
 #define BLK_STS_DM_REQUEUE    ((__force blk_status_t)11)
 
 #define BLK_STS_AGAIN		((__force blk_status_t)12)
+#define BLK_STS_DEV_RESOURCE	((__force blk_status_t)13)
+#define BLK_STS_ZONE_OPEN_RESOURCE	((__force blk_status_t)14)
+#define BLK_STS_ZONE_ACTIVE_RESOURCE	((__force blk_status_t)15)
+#define BLK_STS_OFFLINE		((__force blk_status_t)16)
+#define BLK_STS_DURATION_LIMIT	((__force blk_status_t)17)
+#define BLK_STS_INVAL	((__force blk_status_t)19)
 
 #define BIO_INLINE_VECS 4
 
@@ -129,15 +141,13 @@ struct bio {
 	struct bio_vec		*bi_io_vec;	/* the actual vec list */
 
 	struct bio_set		*bi_pool;
-
-	/*
-	 * We can inline a number of vecs at the end of the bio, to avoid
-	 * double allocations for a small number of bio_vecs. This member
-	 * MUST obviously be kept at the very end of the bio.
-	 */
-	struct bio_vec		bi_inline_vecs[0];
 };
 
+static inline struct bio_vec *bio_inline_vecs(struct bio *bio)
+{
+	return (struct bio_vec *)(bio + 1);
+}
+
 #define BIO_RESET_BYTES		offsetof(struct bio, bi_max_vecs)
 
 /*
@@ -259,6 +269,7 @@ static inline void bio_set_op_attrs(struct bio *bio, unsigned op,
 
 #define REQ_FUA			(1ULL << __REQ_FUA)
 #define REQ_PREFLUSH		(1ULL << __REQ_PREFLUSH)
+#define REQ_IDLE		(1ULL << __REQ_IDLE)
 
 #define RW_MASK			REQ_OP_WRITE
 
