@@ -18,6 +18,8 @@
 
 #include "init/error.h"
 
+#include "sb/counters.h"
+
 #include "util/enumerated_ref.h"
 
 #include <linux/prefetch.h>
@@ -55,12 +57,14 @@ static inline bool wb_key_ref_cmp(const struct wb_key_ref *l, const struct wb_ke
 #endif
 }
 
-static int wb_key_seq_cmp(const void *_l, const void *_r)
+static int wb_key_seq_cmp(const void *_l, const void *_r, const void *priv)
 {
-	const struct btree_write_buffered_key *l = _l;
-	const struct btree_write_buffered_key *r = _r;
+	const struct btree_write_buffer_keys *keys = priv;
+	const struct wb_key_ref *l = _l;
+	const struct wb_key_ref *r = _r;
 
-	return cmp_int(l->journal_seq, r->journal_seq);
+	return cmp_int(wb_keys_idx(keys, l->idx)->journal_seq,
+		       wb_keys_idx(keys, r->idx)->journal_seq);
 }
 
 /* Compare excluding idx, the low 24 bits: */
@@ -135,6 +139,7 @@ static noinline int wb_flush_one_slowpath(struct btree_trans *trans,
 				  BCH_TRANS_COMMIT_no_enospc|
 				  BCH_TRANS_COMMIT_no_check_rw|
 				  BCH_TRANS_COMMIT_no_journal_res|
+				  BCH_TRANS_COMMIT_no_skip_noops|
 				  BCH_TRANS_COMMIT_journal_reclaim);
 }
 
@@ -142,13 +147,13 @@ static inline int wb_flush_one(struct btree_trans *trans, struct btree_iter *ite
 			       struct btree_write_buffered_key *wb,
 			       bool *write_locked,
 			       bool *accounting_accumulated,
-			       size_t *fast)
+			       size_t *fast, size_t *noop)
 {
 	struct btree_path *path;
 
 	EBUG_ON(!wb->journal_seq);
-	EBUG_ON(!trans->c->btree_write_buffer.flushing.pin.seq);
-	EBUG_ON(trans->c->btree_write_buffer.flushing.pin.seq > wb->journal_seq);
+	EBUG_ON(!trans->c->btree.write_buffer.flushing.pin.seq);
+	EBUG_ON(trans->c->btree.write_buffer.flushing.pin.seq > wb->journal_seq);
 
 	try(bch2_btree_iter_traverse(iter));
 
@@ -157,8 +162,9 @@ static inline int wb_flush_one(struct btree_trans *trans, struct btree_iter *ite
 		struct bkey_s_c k = bch2_btree_path_peek_slot_exact(btree_iter_path(trans, iter), &u);
 
 		if (k.k->type == KEY_TYPE_accounting)
-			bch2_accounting_accumulate(bkey_i_to_accounting(&wb->k),
-						   bkey_s_c_to_accounting(k));
+			bch2_accounting_accumulate_maybe_kill(trans->c,
+					bkey_i_to_accounting(&wb->k),
+					bkey_s_c_to_accounting(k));
 	}
 	*accounting_accumulated = true;
 
@@ -171,14 +177,31 @@ static inline int wb_flush_one(struct btree_trans *trans, struct btree_iter *ite
 
 	path = btree_iter_path(trans, iter);
 
-	if (!*write_locked) {
-		try(bch2_btree_node_lock_write(trans, path, &path->l[0].b->c));
+	struct btree_path_level *l = path_l(path);
+	struct bkey_packed *old_p = bch2_btree_node_iter_peek_all(&l->iter, l->b);
+	if (old_p && bkey_cmp_left_packed(l->b, old_p, &wb->k.k.p))
+		old_p = NULL;
 
-		bch2_btree_node_prep_for_write(trans, path, path->l[0].b);
+	struct bkey old_u;
+	struct bkey_s_c old = old_p
+		? bkey_disassemble(l->b, old_p, &old_u)
+		: bkey_s_c_null;
+
+	if (old.k && bkey_and_val_eq(old, bkey_i_to_s_c(&wb->k))) {
+		(*noop)++;
+		return 0;
+	}
+
+	struct btree *b = path->l[0].b;
+
+	if (!*write_locked) {
+		try(bch2_btree_node_lock_write(trans, path, &b->c));
+
+		bch2_btree_node_prep_for_write(trans, path, b);
 		*write_locked = true;
 	}
 
-	if (unlikely(!bch2_btree_node_insert_fits(path->l[0].b, wb->k.k.u64s))) {
+	if (unlikely(!bch2_btree_node_insert_fits(b, wb->k.k.u64s))) {
 		*write_locked = false;
 		return wb_flush_one_slowpath(trans, iter, wb);
 	}
@@ -187,6 +210,10 @@ static inline int wb_flush_one(struct btree_trans *trans, struct btree_iter *ite
 
 	bch2_btree_insert_key_leaf(trans, path, &wb->k, wb->journal_seq);
 	(*fast)++;
+
+	if (unlikely(btree_node_needs_merge(trans, b, 0)))
+		bch2_btree_node_merge_async(trans->c, b);
+
 	return 0;
 }
 
@@ -202,7 +229,7 @@ static inline int wb_flush_one(struct btree_trans *trans, struct btree_iter *ite
  */
 static int
 btree_write_buffered_insert(struct btree_trans *trans,
-			  struct btree_write_buffered_key *wb)
+			    struct btree_write_buffered_key *wb)
 {
 	CLASS(btree_iter, iter)(trans, wb->btree, bkey_start_pos(&wb->k.k),
 				BTREE_ITER_cached|BTREE_ITER_intent);
@@ -214,15 +241,15 @@ btree_write_buffered_insert(struct btree_trans *trans,
 				  BTREE_UPDATE_internal_snapshot_node);
 }
 
-static void move_keys_from_inc_to_flushing(struct btree_write_buffer *wb)
+static void move_keys_from_inc_to_flushing(struct bch_fs_btree_write_buffer *wb)
 {
-	struct bch_fs *c = container_of(wb, struct bch_fs, btree_write_buffer);
+	struct bch_fs *c = container_of(wb, struct bch_fs, btree.write_buffer);
 	struct journal *j = &c->journal;
 
 	if (!wb->inc.keys.nr)
 		return;
 
-	bch2_journal_pin_add(j, wb->inc.keys.data[0].journal_seq, &wb->flushing.pin,
+	bch2_journal_pin_add(j, wb_keys_start(&wb->inc)->journal_seq, &wb->flushing.pin,
 			     bch2_btree_write_buffer_journal_flush);
 
 	darray_resize(&wb->flushing.keys, min_t(size_t, 1U << 20, wb->flushing.keys.nr + wb->inc.keys.nr));
@@ -251,7 +278,7 @@ out:
 	if (!wb->inc.keys.nr)
 		bch2_journal_pin_drop(j, &wb->inc.pin);
 	else
-		bch2_journal_pin_update(j, wb->inc.keys.data[0].journal_seq, &wb->inc.pin,
+		bch2_journal_pin_update(j, wb_keys_start(&wb->inc)->journal_seq, &wb->inc.pin,
 					bch2_btree_write_buffer_journal_flush);
 
 	if (j->watermark) {
@@ -280,9 +307,9 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 {
 	struct bch_fs *c = trans->c;
 	struct journal *j = &c->journal;
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 	struct btree_iter iter = { NULL };
-	size_t overwritten = 0, fast = 0, slowpath = 0, could_not_insert = 0;
+	size_t overwritten = 0, fast = 0, noop = 0, slowpath = 0, could_not_insert = 0;
 	bool write_locked = false;
 	bool accounting_replay_done = test_bit(BCH_FS_accounting_replay_done, &c->flags);
 	int ret = 0;
@@ -301,12 +328,19 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 	u64 start_time = local_clock();
 	u64 nr_flushing = wb->flushing.keys.nr;
 
-	for (size_t i = 0; i < wb->flushing.keys.nr; i++) {
-		wb->sorted.data[i].idx = i;
-		wb->sorted.data[i].btree = wb->flushing.keys.data[i].btree;
-		memcpy(&wb->sorted.data[i].pos, &wb->flushing.keys.data[i].k.k.p, sizeof(struct bpos));
+	wb->sorted.nr = 0;
+	wb_keys_for_each(&wb->flushing, k) {
+		if (wb->sorted.nr == wb->sorted.size)
+			panic("Overflowed wb->sorted at %zu, flushing size %zu\n",
+			      wb->sorted.nr, wb->flushing.keys.nr);
+
+		struct wb_key_ref *dst = &darray_top(wb->sorted);
+		wb->sorted.nr++;
+
+		dst->idx	= (u64 *) k - wb->flushing.keys.data;
+		dst->btree	= k->btree;
+		memcpy(&dst->pos, &k->k.k.p, sizeof(struct bpos));
 	}
-	wb->sorted.nr = wb->flushing.keys.nr;
 
 	/*
 	 * We first sort so that we can detect and skip redundant updates, and
@@ -325,7 +359,7 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 	wb_sort(wb->sorted.data, wb->sorted.nr);
 
 	darray_for_each(wb->sorted, i) {
-		struct btree_write_buffered_key *k = &wb->flushing.keys.data[i->idx];
+		struct btree_write_buffered_key *k = wb_keys_idx(&wb->flushing, i->idx);
 
 		ret = bch2_btree_write_buffer_insert_checks(c, k->btree, &k->k);
 		if (unlikely(ret))
@@ -344,7 +378,7 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 
 		if (i + 1 < &darray_top(wb->sorted) &&
 		    wb_key_eq(i, i + 1)) {
-			struct btree_write_buffered_key *n = &wb->flushing.keys.data[i[1].idx];
+			struct btree_write_buffered_key *n = wb_keys_idx(&wb->flushing, i[1].idx);
 
 			if (k->k.k.type == KEY_TYPE_accounting &&
 			    n->k.k.type == KEY_TYPE_accounting)
@@ -364,24 +398,12 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 			    bpos_gt(k->k.k.p, path->l[0].b->key.k.p)) {
 				bch2_btree_node_unlock_write(trans, path, path->l[0].b);
 				write_locked = false;
-
-				ret = lockrestart_do(trans,
-					bch2_btree_iter_traverse(&iter) ?:
-					bch2_foreground_maybe_merge(trans, iter.path, 0,
-							BCH_WATERMARK_reclaim|
-							BCH_TRANS_COMMIT_journal_reclaim|
-							BCH_TRANS_COMMIT_no_check_rw|
-							BCH_TRANS_COMMIT_no_enospc));
-				if (ret)
-					goto err;
 			}
 		}
 
-		if (!iter.path || iter.btree_id != k->btree) {
-			bch2_trans_iter_exit(&iter);
+		if (!iter.path || iter.btree_id != k->btree)
 			bch2_trans_iter_init(trans, &iter, k->btree, k->k.k.p,
 					     BTREE_ITER_intent|BTREE_ITER_all_snapshots);
-		}
 
 		bch2_btree_iter_set_pos(&iter, k->k.k.p);
 		btree_iter_path(trans, &iter)->preserve = false;
@@ -394,7 +416,7 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 			}
 
 			ret = wb_flush_one(trans, &iter, k, &write_locked,
-					   &accounting_accumulated, &fast);
+					   &accounting_accumulated, &fast, &noop);
 			if (!write_locked)
 				bch2_trans_begin(trans);
 		} while (bch2_err_matches(ret, BCH_ERR_transaction_restart));
@@ -424,25 +446,28 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 		 * The fastpath zapped the seq of keys that were successfully flushed so
 		 * we can skip those here.
 		 */
-		trace_and_count(c, write_buffer_flush_slowpath, trans, slowpath, wb->flushing.keys.nr);
+		event_inc_trace(c, write_buffer_flush_slowpath, buf,
+				prt_printf(&buf, "%zu/%zu", slowpath, wb->flushing.keys.nr));
 
-		sort_nonatomic(wb->flushing.keys.data,
-			       wb->flushing.keys.nr,
-			       sizeof(wb->flushing.keys.data[0]),
-			       wb_key_seq_cmp, NULL);
+		sort_r_nonatomic(wb->sorted.data,
+				 wb->sorted.nr,
+				 sizeof(wb->sorted.data[0]),
+				 wb_key_seq_cmp, NULL,
+				 &wb->flushing);
 
-		darray_for_each(wb->flushing.keys, i) {
-			if (!i->journal_seq)
+		darray_for_each(wb->sorted, i) {
+			struct btree_write_buffered_key *k = wb_keys_idx(&wb->flushing, i->idx);
+			if (!k->journal_seq)
 				continue;
 
 			if (!accounting_replay_done &&
-			    i->k.k.type == KEY_TYPE_accounting) {
+			    k->k.k.type == KEY_TYPE_accounting) {
 				could_not_insert++;
 				continue;
 			}
 
 			if (!could_not_insert)
-				bch2_journal_pin_update(j, i->journal_seq, &wb->flushing.pin,
+				bch2_journal_pin_update(j, k->journal_seq, &wb->flushing.pin,
 							bch2_btree_write_buffer_journal_flush);
 
 			bch2_trans_begin(trans);
@@ -453,11 +478,11 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 					BCH_TRANS_COMMIT_no_check_rw|
 					BCH_TRANS_COMMIT_no_enospc|
 					BCH_TRANS_COMMIT_no_journal_res ,
-					btree_write_buffered_insert(trans, i));
+					btree_write_buffered_insert(trans, k));
 			if (ret)
 				goto err;
 
-			i->journal_seq = 0;
+			k->journal_seq = 0;
 		}
 
 		/*
@@ -479,12 +504,14 @@ static int bch2_btree_write_buffer_flush_locked(struct btree_trans *trans)
 		 * distinct counters touched somehow was very large.
 		 */
 		if (could_not_insert) {
-			struct btree_write_buffered_key *dst = wb->flushing.keys.data;
+			struct btree_write_buffered_key *dst = wb_keys_start(&wb->flushing);
 
-			darray_for_each(wb->flushing.keys, i)
-				if (i->journal_seq)
-					*dst++ = *i;
-			wb->flushing.keys.nr = dst - wb->flushing.keys.data;
+			wb_keys_for_each_safe(&wb->flushing, i)
+				if (i->journal_seq) {
+					memmove_u64s_down(dst, i, wb_key_u64s(&i->k));
+					dst = wb_key_next(dst);
+				}
+			wb->flushing.keys.nr = (u64 *) dst - wb->flushing.keys.data;
 		}
 	}
 err:
@@ -493,9 +520,13 @@ err:
 		wb->flushing.keys.nr = 0;
 	}
 
-	bch2_time_stats_update(&c->times[BCH_TIME_btree_write_buffer_flush], start_time);
 	bch2_fs_fatal_err_on(ret, c, "%s", bch2_err_str(ret));
-	trace_write_buffer_flush(trans, nr_flushing, overwritten, fast);
+
+	bch2_time_stats_update(&c->times[BCH_TIME_btree_write_buffer_flush], start_time);
+	event_inc_trace(c, write_buffer_flush, buf,
+		prt_printf(&buf, "flushed %llu skipped %zu fast %zu noop %zu",
+			   nr_flushing, overwritten, fast, noop));
+
 	return ret;
 }
 
@@ -550,7 +581,7 @@ static int btree_write_buffer_flush_seq(struct btree_trans *trans, u64 max_seq,
 					bool *did_work)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 	int ret = 0, fetch_from_journal_err;
 
 	do {
@@ -589,7 +620,7 @@ int bch2_btree_write_buffer_flush_sync(struct btree_trans *trans)
 	struct bch_fs *c = trans->c;
 	bool did_work = false;
 
-	trace_and_count(c, write_buffer_flush_sync, trans, _RET_IP_);
+	event_inc_trace(c, write_buffer_flush_sync, buf, prt_str(&buf, trans->fn));
 
 	return btree_write_buffer_flush_seq(trans, journal_cur_seq(&c->journal), &did_work);
 }
@@ -612,7 +643,7 @@ bool bch2_btree_write_buffer_flush_going_ro(struct bch_fs *c)
 int bch2_btree_write_buffer_flush_nocheck_rw(struct btree_trans *trans)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 	int ret = 0;
 
 	if (mutex_trylock(&wb->flushing.lock)) {
@@ -642,17 +673,20 @@ int bch2_btree_write_buffer_tryflush(struct btree_trans *trans)
  */
 int bch2_btree_write_buffer_maybe_flush(struct btree_trans *trans,
 					struct bkey_s_c referring_k,
-					struct bkey_buf *last_flushed)
+					struct wb_maybe_flush *f)
 {
 	struct bch_fs *c = trans->c;
 
-	if (!bkey_and_val_eq(referring_k, bkey_i_to_s_c(last_flushed->k))) {
-		if (trace_write_buffer_maybe_flush_enabled()) {
-			CLASS(printbuf, buf)();
+	if (f->seen_error &&
+	    f->nr_flushes > 32 &&
+	    f->nr_flushes * 8 > f->nr_done)
+		return 0;
 
+	if (!bkey_and_val_eq(referring_k, bkey_i_to_s_c(f->last_flushed.k))) {
+		event_inc_trace(c, write_buffer_maybe_flush, buf, ({
+			prt_printf(&buf, "%s\n", trans->fn);
 			bch2_bkey_val_to_text(&buf, c, referring_k);
-			trace_write_buffer_maybe_flush(trans, _RET_IP_, buf.buf);
-		}
+		}));
 
 		struct bkey_buf tmp __cleanup(bch2_bkey_buf_exit);
 		bch2_bkey_buf_init(&tmp);
@@ -663,22 +697,25 @@ int bch2_btree_write_buffer_maybe_flush(struct btree_trans *trans,
 			bch2_btree_interior_updates_flush(c);
 		}
 
-		try(bch2_btree_write_buffer_flush_sync(trans));
+		bool did_work = false;
+		try(btree_write_buffer_flush_seq(trans, journal_cur_seq(&c->journal), &did_work));
 
-		bch2_bkey_buf_copy(last_flushed, tmp.k);
+		bch2_bkey_buf_copy(&f->last_flushed, tmp.k);
+		f->nr_flushes++;
 
 		/* can we avoid the unconditional restart? */
-		trace_and_count(c, trans_restart_write_buffer_flush, trans, _RET_IP_);
+		event_inc_trace(c, trans_restart_write_buffer_flush, buf, prt_str(&buf, trans->fn));
 		return bch_err_throw(c, transaction_restart_write_buffer_flush);
 	}
 
+	f->seen_error = true;
 	return 0;
 }
 
 static void bch2_btree_write_buffer_flush_work(struct work_struct *work)
 {
-	struct bch_fs *c = container_of(work, struct bch_fs, btree_write_buffer.flush_work);
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs *c = container_of(work, struct bch_fs, btree.write_buffer.flush_work);
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 	int ret;
 
 	scoped_guard(mutex, &wb->flushing.lock) {
@@ -691,7 +728,7 @@ static void bch2_btree_write_buffer_flush_work(struct work_struct *work)
 	enumerated_ref_put(&c->writes, BCH_WRITE_REF_btree_write_buffer);
 }
 
-static void wb_accounting_sort(struct btree_write_buffer *wb)
+static void wb_accounting_sort(struct bch_fs_btree_write_buffer *wb)
 {
 	eytzinger0_sort(wb->accounting.data, wb->accounting.nr,
 			sizeof(wb->accounting.data[0]),
@@ -701,15 +738,12 @@ static void wb_accounting_sort(struct btree_write_buffer *wb)
 int bch2_accounting_key_to_wb_slowpath(struct bch_fs *c, enum btree_id btree,
 				       struct bkey_i_accounting *k)
 {
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 
-	if (trace_accounting_key_to_wb_slowpath_enabled()) {
-		CLASS(printbuf, buf)();
+	event_inc_trace(c, accounting_key_to_wb_slowpath, buf, ({
 		prt_printf(&buf, "have: %zu\n", wb->accounting.nr);
 		bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&k->k_i));
-		trace_accounting_key_to_wb_slowpath(c, buf.buf);
-	}
-	count_event(c, accounting_key_to_wb_slowpath);
+	}));
 
 	struct btree_write_buffered_key new = { .btree = btree };
 	bkey_copy(&new.k, &k->k_i);
@@ -724,17 +758,18 @@ int bch2_journal_key_to_wb_slowpath(struct bch_fs *c,
 			     struct journal_keys_to_wb *dst,
 			     enum btree_id btree, struct bkey_i *k)
 {
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
+	unsigned u64s = wb_key_u64s(k);
 	int ret;
 retry:
-	ret = darray_make_room_gfp(&dst->wb->keys, 1, GFP_KERNEL);
+	ret = darray_make_room_gfp(&dst->wb->keys, u64s, GFP_KERNEL);
 	if (!ret && dst->wb == &wb->flushing)
 		ret = darray_resize(&wb->sorted, wb->flushing.keys.size);
 
 	if (unlikely(ret)) {
-		if (dst->wb == &c->btree_write_buffer.flushing) {
+		if (dst->wb == &c->btree.write_buffer.flushing) {
 			mutex_unlock(&dst->wb->lock);
-			dst->wb = &c->btree_write_buffer.inc;
+			dst->wb = &c->btree.write_buffer.inc;
 			bch2_journal_pin_add(&c->journal, dst->seq, &dst->wb->pin,
 					     bch2_btree_write_buffer_journal_flush);
 			goto retry;
@@ -746,21 +781,16 @@ retry:
 	dst->room = darray_room(dst->wb->keys);
 	if (dst->wb == &wb->flushing)
 		dst->room = min(dst->room, wb->sorted.size - wb->flushing.keys.nr);
-	BUG_ON(!dst->room);
+	BUG_ON(dst->room < u64s);
 	BUG_ON(!dst->seq);
 
-	struct btree_write_buffered_key *wb_k = &darray_top(dst->wb->keys);
-	wb_k->journal_seq	= dst->seq;
-	wb_k->btree		= btree;
-	bkey_copy(&wb_k->k, k);
-	dst->wb->keys.nr++;
-	dst->room--;
+	bch2_journal_key_to_wb_reserved(c, dst, btree, k);
 	return 0;
 }
 
 void bch2_journal_keys_to_write_buffer_start(struct bch_fs *c, struct journal_keys_to_wb *dst, u64 seq)
 {
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 
 	if (mutex_trylock(&wb->flushing.lock)) {
 		mutex_lock(&wb->inc.lock);
@@ -796,7 +826,7 @@ void bch2_journal_keys_to_write_buffer_start(struct bch_fs *c, struct journal_ke
 
 int bch2_journal_keys_to_write_buffer_end(struct bch_fs *c, struct journal_keys_to_wb *dst)
 {
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 	unsigned live_accounting_keys = 0;
 	int ret = 0;
 
@@ -824,7 +854,7 @@ int bch2_journal_keys_to_write_buffer_end(struct bch_fs *c, struct journal_keys_
 
 	if (bch2_btree_write_buffer_should_flush(c) &&
 	    __enumerated_ref_tryget(&c->writes, BCH_WRITE_REF_btree_write_buffer) &&
-	    !queue_work(system_unbound_wq, &c->btree_write_buffer.flush_work))
+	    !queue_work(system_unbound_wq, &c->btree.write_buffer.flush_work))
 		enumerated_ref_put(&c->writes, BCH_WRITE_REF_btree_write_buffer);
 
 	if (dst->wb == &wb->flushing)
@@ -849,7 +879,7 @@ static int wb_keys_resize(struct btree_write_buffer_keys *wb, size_t new_size)
 
 int bch2_btree_write_buffer_resize(struct bch_fs *c, size_t new_size)
 {
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 
 	return wb_keys_resize(&wb->flushing, new_size) ?:
 		wb_keys_resize(&wb->inc, new_size);
@@ -857,7 +887,7 @@ int bch2_btree_write_buffer_resize(struct bch_fs *c, size_t new_size)
 
 void bch2_fs_btree_write_buffer_exit(struct bch_fs *c)
 {
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 
 	BUG_ON((wb->inc.keys.nr || wb->flushing.keys.nr) &&
 	       !bch2_journal_error(&c->journal));
@@ -870,7 +900,7 @@ void bch2_fs_btree_write_buffer_exit(struct bch_fs *c)
 
 void bch2_fs_btree_write_buffer_init_early(struct bch_fs *c)
 {
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 
 	mutex_init(&wb->inc.lock);
 	mutex_init(&wb->flushing.lock);
@@ -879,7 +909,7 @@ void bch2_fs_btree_write_buffer_init_early(struct bch_fs *c)
 
 int bch2_fs_btree_write_buffer_init(struct bch_fs *c)
 {
-	struct btree_write_buffer *wb = &c->btree_write_buffer;
+	struct bch_fs_btree_write_buffer *wb = &c->btree.write_buffer;
 
 	/* Will be resized by journal as needed: */
 	unsigned initial_size = 1 << 16;
