@@ -382,11 +382,10 @@ out:									\
 
 void bch2_io_failures_to_text(struct printbuf *, struct bch_fs *,
 			      struct bch_io_failures *);
-struct bch_dev_io_failures *bch2_dev_io_failures(struct bch_io_failures *,
-						 unsigned);
-void bch2_mark_io_failure(struct bch_io_failures *,
-			  struct extent_ptr_decoded *, bool);
-void bch2_mark_btree_validate_failure(struct bch_io_failures *, unsigned);
+struct bch_dev_io_failures *bch2_dev_io_failures(struct bch_io_failures *, unsigned);
+struct bch_dev_io_failures *bch2_dev_io_failures_mut(struct bch_io_failures *, unsigned);
+void bch2_mark_io_failure(struct bch_io_failures *, struct extent_ptr_decoded *, int);
+void bch2_mark_dev_io_failure(struct bch_io_failures *, unsigned, int);
 int bch2_bkey_pick_read_device(struct bch_fs *, struct bkey_s_c,
 			       struct bch_io_failures *,
 			       struct extent_ptr_decoded *, int);
@@ -481,6 +480,13 @@ static inline bool bkey_extent_is_direct_data(const struct bkey *k)
 	}
 }
 
+static inline bool bkey_is_user_data(const struct bkey *k)
+{
+	return  k->type == KEY_TYPE_extent ||
+		k->type == KEY_TYPE_inline_data ||
+		k->type == KEY_TYPE_reservation;
+}
+
 static inline bool bkey_extent_is_inline_data(const struct bkey *k)
 {
 	return  k->type == KEY_TYPE_inline_data ||
@@ -568,19 +574,12 @@ unsigned bch2_bkey_sectors_compressed(const struct bch_fs *, struct bkey_s_c);
 
 unsigned bch2_bkey_replicas(struct bch_fs *, struct bkey_s_c);
 
-static inline unsigned __extent_ptr_durability(struct bch_dev *ca, struct extent_ptr_decoded *p)
-{
-	if (p->ptr.cached)
-		return 0;
-
-	return p->has_ec
-		? p->ec.redundancy + 1
-		: ca->mi.durability;
-}
-
-unsigned bch2_extent_ptr_desired_durability(struct bch_fs *, struct extent_ptr_decoded *);
-unsigned bch2_extent_ptr_durability(struct bch_fs *, struct extent_ptr_decoded *);
-unsigned bch2_bkey_durability(struct bch_fs *, struct bkey_s_c);
+unsigned bch2_dev_durability(struct bch_fs *, unsigned);
+int bch2_extent_ptr_desired_durability(struct btree_trans *, struct extent_ptr_decoded *);
+int bch2_extent_ptr_durability(struct btree_trans *, struct extent_ptr_decoded *);
+int bch2_bkey_durability(struct btree_trans *, struct bkey_s_c);
+unsigned bch2_btree_ptr_durability(struct bch_fs *, struct bkey_s_c);
+bool bch2_bkey_can_read(const struct bch_fs *, struct bkey_s_c);
 
 const struct bch_extent_ptr *bch2_bkey_has_device_c(const struct bch_fs *,
 						    struct bkey_s_c, unsigned);
@@ -596,6 +595,7 @@ bool bch2_bkey_devs_rw(struct bch_fs *, struct bkey_s_c);
 bool bch2_bkey_has_target(struct bch_fs *, struct bkey_s_c, unsigned);
 bool bch2_bkey_in_target(struct bch_fs *, struct bkey_s_c, unsigned);
 
+void bch2_bkey_extent_entry_drop_s(const struct bch_fs *, struct bkey_s, union bch_extent_entry *);
 void bch2_bkey_extent_entry_drop(const struct bch_fs *, struct bkey_i *, union bch_extent_entry *);
 
 static inline void bch2_bkey_append_ptr(const struct bch_fs *c, struct bkey_i *k, struct bch_extent_ptr ptr)
@@ -624,9 +624,11 @@ void bch2_extent_ptr_decoded_append(const struct bch_fs *, struct bkey_i *,
 				    struct extent_ptr_decoded *);
 void bch2_bkey_drop_ptr_noerror(const struct bch_fs *, struct bkey_s, struct bch_extent_ptr *);
 void bch2_bkey_drop_ptr(const struct bch_fs *, struct bkey_s, struct bch_extent_ptr *);
+void bch2_bkey_drop_ptrs_mask(const struct bch_fs *, struct bkey_i *, unsigned);
 
+void bch2_bkey_drop_device_noerror(const struct bch_fs *, struct bkey_s, unsigned);
 void bch2_bkey_drop_device(const struct bch_fs *, struct bkey_s, unsigned);
-void bch2_bkey_drop_ec(const struct bch_fs *, struct bkey_i *k, unsigned);
+void bch2_bkey_drop_ec_mask(const struct bch_fs *, struct bkey_i *k, unsigned);
 
 #define bch2_bkey_drop_ptrs_noerror(_k, _p, _entry, _cond)			\
 do {										\
@@ -677,7 +679,7 @@ void bch2_extent_ptr_set_cached(struct bch_fs *, struct bch_inode_opts *,
 
 int bch2_bkey_drop_stale_ptrs(struct btree_trans *, struct btree_iter *, struct bkey_s_c);
 void bch2_bkey_drop_extra_cached_ptrs(struct bch_fs *, struct bch_inode_opts *, struct bkey_s);
-void bch2_bkey_drop_extra_durability(struct bch_fs *, struct bch_inode_opts *, struct bkey_s);
+int bch2_bkey_drop_extra_durability(struct btree_trans *, struct bch_inode_opts *, struct bkey_s);
 
 void bch2_extent_ptr_to_text(struct printbuf *out, struct bch_fs *, const struct bch_extent_ptr *);
 void bch2_bkey_ptrs_to_text(struct printbuf *, struct bch_fs *,
