@@ -185,11 +185,6 @@ static inline int should_promote(struct bch_fs *c, struct bkey_s_c k,
 	if (!self_healing) {
 		BUG_ON(!opts.promote_target);
 
-		if (!(flags & BCH_READ_may_promote)) {
-			event_inc(c, data_read_nopromote_may_not);
-			return bch_err_throw(c, nopromote_may_not);
-		}
-
 		if (bch2_bkey_has_target(c, k, opts.promote_target)) {
 			event_inc(c, data_read_nopromote_already_promoted);
 			return bch_err_throw(c, nopromote_already_promoted);
@@ -901,12 +896,15 @@ static void __bch2_read_endio(struct work_struct *work)
 	if (unlikely(rbio->narrow_crcs))
 		bch2_rbio_narrow_crcs(rbio);
 
-	if (likely(!parent->data_update)) {
+	if (likely(!parent->data_update) ||
+	    parent->data_update_verify_decompress) {
 		/* Adjust crc to point to subset of data we want: */
 		crc.offset     += rbio->offset_into_extent;
 		crc.live_size	= bvec_iter_sectors(rbio->bvec_iter);
 
 		if (crc_is_compressed(crc)) {
+			BUG_ON(!rbio->bounce);
+
 			ret = bch2_encrypt_bio(c, crc.csum_type, nonce, src);
 			if (ret) {
 				bch2_rbio_error(rbio, bch_err_throw(c, data_read_decrypt_err));
@@ -918,7 +916,16 @@ static void __bch2_read_endio(struct work_struct *work)
 				bch2_rbio_error(rbio, bch_err_throw(c, data_read_decompress_err));
 				return;
 			}
-		} else {
+
+			if (unlikely(parent->data_update)) {
+				/* We decrypted to decompress; re-encrypt: */
+				ret = bch2_encrypt_bio(c, crc.csum_type, nonce, src);
+				if (ret) {
+					bch2_rbio_error(rbio, bch_err_throw(c, data_read_decrypt_err));
+					return;
+				}
+			}
+		} else if (likely(!parent->data_update)) {
 			/* don't need to decrypt the entire bio: */
 			nonce = nonce_add(nonce, crc.offset << 9);
 			bio_advance(src, crc.offset << 9);
@@ -938,7 +945,9 @@ static void __bch2_read_endio(struct work_struct *work)
 				bio_copy_data_iter(dst, &dst_iter, src, &src_iter);
 			}
 		}
-	} else {
+	}
+
+	if (unlikely(parent->data_update)) {
 		if (rbio->split)
 			rbio->parent->pick = rbio->pick;
 
@@ -1073,7 +1082,9 @@ static inline struct bch_read_bio *read_extent_rbio_alloc(struct btree_trans *tr
 	struct bch_fs *c = trans->c;
 	struct bpos data_pos = bkey_start_pos(k.k);
 
-	struct bch_read_bio *rbio = orig->opts.promote_target || have_io_error(failed)
+	struct bch_read_bio *rbio =
+		(orig->opts.promote_target && (flags & BCH_READ_may_promote)) ||
+		have_io_error(failed)
 		? promote_alloc(trans, iter, k, &pick, flags, orig,
 				&bounce, &read_full, failed)
 		: NULL;
@@ -1333,8 +1344,7 @@ int __bch2_read_extent(struct btree_trans *trans,
 {
 	struct bch_fs *c = trans->c;
 	struct extent_ptr_decoded pick;
-	bool bounce = false, read_full = false, narrow_crcs = false;
-	struct data_update *u = rbio_data_update(orig);
+	bool bounce = false, read_full = false;
 	int ret = 0;
 
 	if (bkey_extent_is_inline_data(k.k))
@@ -1379,18 +1389,43 @@ int __bch2_read_extent(struct btree_trans *trans,
 		return read_extent_done(orig, flags, bch_err_throw(c, data_read_ptr_stale_dirty));
 	}
 
-	if (likely(!u)) {
-		if (!(flags & BCH_READ_last_fragment) ||
-		    bio_flagged(&orig->bio, BIO_CHAIN))
-			flags |= BCH_READ_must_clone;
+	if (!(flags & BCH_READ_last_fragment) ||
+	    bio_flagged(&orig->bio, BIO_CHAIN))
+		flags |= BCH_READ_must_clone;
 
-		narrow_crcs = !(flags & BCH_READ_in_retry) && can_narrow_crc(pick.crc);
+	bool narrow_crcs = !orig->data_update &&
+		!(flags & BCH_READ_in_retry) &&
+		can_narrow_crc(pick.crc);
 
-		if (narrow_crcs && (flags & BCH_READ_user_mapped))
-			flags |= BCH_READ_must_bounce;
+	if (narrow_crcs && (flags & BCH_READ_user_mapped))
+		flags |= BCH_READ_must_bounce;
 
+
+	if (likely(!orig->data_update)) {
+		/* Check we're not trying to read more than we have in this extent: */
 		EBUG_ON(offset_into_extent + bvec_iter_sectors(iter) > k.k->size);
+	} else {
+		/*
+		 * For data updates, iter.bi_size can be bigger than the key - it's just
+		 * the buffer size, we're not trying to read a specific amount of data
+		 * from this extent:
+		 *
+		 * If it's smaller, we raced with a merge during a retry:
+		 */
+		if (unlikely(pick.crc.uncompressed_size > bvec_iter_sectors(iter))) {
+			if (ca)
+				enumerated_ref_put(&ca->io_ref[READ],
+					BCH_DEV_READ_REF_io_read);
+			return read_extent_done(orig, flags, bch_err_throw(c, data_read_buffer_too_small));
+		}
 
+		iter.bi_size = pick.crc.compressed_size << 9;
+		read_full = true;
+	}
+
+	bool decode = !orig->data_update || orig->data_update_verify_decompress;
+
+	if (likely(decode))
 		if (crc_is_compressed(pick.crc) ||
 		    (pick.crc.csum_type != BCH_CSUM_none &&
 		     (bvec_iter_sectors(iter) != pick.crc.uncompressed_size ||
@@ -1400,21 +1435,6 @@ int __bch2_read_extent(struct btree_trans *trans,
 			read_full = true;
 			bounce = true;
 		}
-	} else {
-		/*
-		 * can happen if we retry, and the extent we were going to read
-		 * has been merged in the meantime:
-		 */
-		if (unlikely(pick.crc.compressed_size > u->op.wbio.bio.bi_iter.bi_size)) {
-			if (ca)
-				enumerated_ref_put(&ca->io_ref[READ],
-					BCH_DEV_READ_REF_io_read);
-			return read_extent_done(orig, flags, bch_err_throw(c, data_read_buffer_too_small));
-		}
-
-		iter.bi_size	= pick.crc.compressed_size << 9;
-		read_full = true;
-	}
 
 	struct bch_read_bio *rbio =
 		read_extent_rbio_alloc(trans, orig, iter, read_pos, data_btree, k,
