@@ -517,6 +517,15 @@ static int reconstruct_inode(struct btree_trans *trans, enum btree_id btree, u32
 	new_inode.bi_inum = inum;
 	new_inode.bi_snapshot = snapshot;
 
+	struct bch_inode_unpacked ancestor;
+	int ret = bch2_inode_find_oldest_snapshot(trans, inum, snapshot, &ancestor);
+	if (ret && !bch2_err_matches(ret, ENOENT))
+		return ret;
+	if (!ret) {
+		new_inode.bi_hash_seed = ancestor.bi_hash_seed;
+		SET_INODE_STR_HASH(&new_inode, INODE_STR_HASH(&ancestor));
+	}
+
 	return __bch2_fsck_write_inode(trans, &new_inode);
 }
 
@@ -721,11 +730,11 @@ lookup_inode_for_snapshot(struct btree_trans *trans, struct inode_walker *w, str
 		}
 
 		if (ret)
-			goto fsck_err;
+			return ERR_PTR(ret);
 
 		ret = bch2_trans_commit(trans, NULL, NULL, 0);
 		if (ret)
-			goto fsck_err;
+			return ERR_PTR(ret);
 
 		struct inode_walker_entry new_entry = *i;
 
@@ -739,10 +748,9 @@ lookup_inode_for_snapshot(struct btree_trans *trans, struct inode_walker *w, str
 		size_t pos = i - w->inodes.data;
 		ret = darray_insert_item(&w->inodes, pos, new_entry);
 		if (ret)
-			goto fsck_err;
+			return ERR_PTR(ret);
 
-		ret = bch_err_throw(c, transaction_restart_nested);
-		goto fsck_err;
+		return ERR_PTR(bch_err_throw(c, transaction_restart_nested));
 	}
 
 	return i;
@@ -887,51 +895,37 @@ static int check_inode(struct btree_trans *trans,
 	CLASS(printbuf, buf)();
 	struct bch_inode_unpacked u;
 	bool do_update = false;
-	int ret;
 
-	ret = bch2_check_key_has_snapshot(trans, iter, k);
+	int ret = bch2_check_key_has_snapshot(trans, iter, k);
 	if (ret < 0)
-		goto err;
+		return ret;
 	if (ret)
 		return 0;
 
-	ret = bch2_snapshots_seen_update(c, s, iter->btree_id, k.k->p);
-	if (ret)
-		goto err;
+	try(bch2_snapshots_seen_update(c, s, iter->btree_id, k.k->p));
 
 	if (!bkey_is_inode(k.k))
 		return 0;
 
-	ret = bch2_inode_unpack(k, &u);
-	if (ret)
-		goto err;
+	try(bch2_inode_unpack(k, &u));
 
-	if (snapshot_root->bi_inum != u.bi_inum) {
-		ret = bch2_inode_find_snapshot_root(trans, u.bi_inum, snapshot_root);
-		if (ret)
-			goto err;
-	}
+	if (snapshot_root->bi_inum != u.bi_inum ||
+	    !bch2_snapshot_is_ancestor(c, u.bi_snapshot, snapshot_root->bi_snapshot))
+		try(bch2_inode_find_oldest_snapshot(trans, u.bi_inum, u.bi_snapshot, snapshot_root));
 
 	if (u.bi_hash_seed	!= snapshot_root->bi_hash_seed ||
-	    INODE_STR_HASH(&u)	!= INODE_STR_HASH(snapshot_root)) {
-		ret = bch2_repair_inode_hash_info(trans, snapshot_root);
-		BUG_ON(ret == -BCH_ERR_fsck_repair_unimplemented);
-		if (ret)
-			goto err;
-	}
+	    INODE_STR_HASH(&u)	!= INODE_STR_HASH(snapshot_root))
+		try(bch2_repair_inode_hash_info(trans, &u, snapshot_root));
 
 	ret = bch2_check_inode_has_case_insensitive(trans, &u, &s->ids, &do_update);
 	if (bch2_err_matches(ret, ENOENT)) /* disconnected inode; will be fixed by a later pass */
 		ret = 0;
 	bch_err_msg(c, ret, "bch2_check_inode_has_case_insensitive()");
 	if (ret)
-		goto err;
+		return ret;
 
-	if (bch2_inode_has_backpointer(&u)) {
-		ret = check_inode_dirent_inode(trans, &u, &do_update);
-		if (ret)
-			goto err;
-	}
+	if (bch2_inode_has_backpointer(&u))
+		try(check_inode_dirent_inode(trans, &u, &do_update));
 
 	if (fsck_err_on(bch2_inode_has_backpointer(&u) &&
 			(u.bi_flags & BCH_INODE_unlinked),
@@ -949,7 +943,7 @@ static int check_inode(struct btree_trans *trans,
 
 		ret = bch2_empty_dir_snapshot(trans, k.k->p.offset, 0, k.k->p.snapshot);
 		if (ret && ret != -BCH_ERR_ENOTEMPTY_dir_not_empty)
-			goto err;
+			return ret;
 
 		fsck_err_on(ret, trans, inode_dir_unlinked_but_not_empty,
 			    "dir unlinked but not empty\n%s",
@@ -971,7 +965,7 @@ static int check_inode(struct btree_trans *trans,
 
 	ret = bch2_inode_has_child_snapshots(trans, k.k->p);
 	if (ret < 0)
-		goto err;
+		return ret;
 
 	if (fsck_err_on(ret != !!(u.bi_flags & BCH_INODE_has_child_snapshot),
 			trans, inode_has_child_snapshots_wrong,
@@ -1010,13 +1004,12 @@ static int check_inode(struct btree_trans *trans,
 				    "inode %llu:%u unlinked, but not on deleted list",
 				    u.bi_inum, k.k->p.snapshot);
 
-			ret = bch2_btree_bit_mod_buffered(trans, BTREE_ID_deleted_inodes, k.k->p, 1);
-			if (ret)
-				goto err;
+			try(bch2_btree_bit_mod_buffered(trans, BTREE_ID_deleted_inodes, k.k->p, 1));
+			ret = 0;
 		} else {
 			ret = bch2_inode_or_descendents_is_open(trans, k.k->p);
 			if (ret < 0)
-				goto err;
+				return ret;
 
 			if (fsck_err_on(!ret,
 					trans, inode_unlinked_and_not_open,
@@ -1045,7 +1038,7 @@ static int check_inode(struct btree_trans *trans,
 
 		ret = bch2_subvolume_get(trans, u.bi_subvol, false, &s);
 		if (ret && !bch2_err_matches(ret, ENOENT))
-			goto err;
+			return ret;
 
 		if (ret && (c->sb.btrees_lost_data & BIT_ULL(BTREE_ID_subvolumes))) {
 			ret = reconstruct_subvol(trans, k.k->p.snapshot, u.bi_subvol, u.bi_inum);
@@ -1087,7 +1080,6 @@ do_update:
 		if (ret)
 			return ret;
 	}
-err:
 fsck_err:
 	bch_err_fn(c, ret);
 	return ret;
@@ -1215,126 +1207,147 @@ static inline bool btree_matches_i_mode(enum btree_id btree, unsigned mode)
 	}
 }
 
+static int count_inode_keys(struct btree_trans *trans,
+			    struct bpos inode_pos,
+			    enum btree_id btree,
+			    struct printbuf *out)
+{
+	struct bkey_s_c k;
+	unsigned nr_keys = 0;
+	int ret = 0;
+	for_each_btree_key_max_norestart(trans, iter, btree,
+					 inode_pos,
+					 POS(inode_pos.inode, U64_MAX),
+					 0, k, ret) {
+		if (k.k->type == KEY_TYPE_error ||
+		    k.k->type == KEY_TYPE_hash_whiteout)
+			continue;
+
+		nr_keys++;
+		if (out && nr_keys <= 10) {
+			bch2_bkey_val_to_text(out, trans->c, k);
+			prt_newline(out);
+		}
+		if (nr_keys >= 100)
+			break;
+	}
+
+	return ret ?: nr_keys;
+}
+
 int bch2_check_key_has_inode(struct btree_trans *trans,
 			     struct btree_iter *iter,
 			     struct inode_walker *inode,
 			     struct inode_walker_entry *i,
 			     struct bkey_s_c k)
 {
-	struct bch_fs *c = trans->c;
-	CLASS(printbuf, buf)();
-	int ret = PTR_ERR_OR_ZERO(i);
-	if (ret)
-		return ret;
+	errptr_try(i);
 
 	if (bkey_extent_whiteout(k.k))
 		return 0;
 
 	bool have_inode = i && !i->whiteout;
 
-	if (!have_inode && (c->sb.btrees_lost_data & BIT_ULL(BTREE_ID_inodes)))
-		goto reconstruct;
-
 	if (have_inode && btree_matches_i_mode(iter->btree_id, i->inode.bi_mode))
 		return 0;
 
-	prt_printf(&buf, ", ");
+	struct bch_fs *c = trans->c;
+	CLASS(printbuf, buf)();
 
-	bool have_old_inode = false;
+	if (have_inode)
+		prt_printf(&buf, "key for wrong inode mode %o", i->inode.bi_mode);
+	else
+		prt_str(&buf, "key in missing inode");
+
+	struct inode_walker_entry *good_ancestor = NULL;
 	darray_for_each(inode->inodes, i2)
 		if (!i2->whiteout &&
 		    bch2_snapshot_is_ancestor(c, k.k->p.snapshot, i2->inode.bi_snapshot) &&
 		    btree_matches_i_mode(iter->btree_id, i2->inode.bi_mode)) {
-			prt_printf(&buf, "but found good inode in older snapshot\n");
+			prt_printf(&buf, ", but found good inode in older snapshot");
 			bch2_inode_unpacked_to_text(&buf, &i2->inode);
 			prt_newline(&buf);
-			have_old_inode = true;
+			good_ancestor = i2;
 			break;
 		}
 
-	struct bkey_s_c k2;
-	unsigned nr_keys = 0;
+	prt_printf(&buf, "\nfound keys:\n");
 
-	prt_printf(&buf, "found keys:\n");
+	struct bpos inode_pos = SPOS(k.k->p.inode, 0, k.k->p.snapshot);
+	int ret = count_inode_keys(trans, inode_pos, iter->btree_id, &buf);
+	if (ret < 0)
+		return ret;
 
-	for_each_btree_key_max_norestart(trans, iter2, iter->btree_id,
-					 SPOS(k.k->p.inode, 0, k.k->p.snapshot),
-					 POS(k.k->p.inode, U64_MAX),
-					 0, k2, ret) {
-		if (k.k->type == KEY_TYPE_error ||
-		    k.k->type == KEY_TYPE_hash_whiteout)
-			continue;
-
-		nr_keys++;
-		if (nr_keys <= 10) {
-			bch2_bkey_val_to_text(&buf, c, k2);
-			prt_newline(&buf);
-		}
-		if (nr_keys >= 100)
-			break;
+	unsigned nr_keys = ret;
+	if (!nr_keys) {
+		bch_err(c, "%s: error finding live keys in inode", __func__);
+		return bch_err_throw(c, shutdown_with_errors_unfixed);
 	}
 
-	if (ret)
-		goto err;
-
-	unsigned reconstruct_limit = iter->btree_id == BTREE_ID_extents ? 3 : 0;
-
 	if (nr_keys > 100)
-		prt_printf(&buf, "found > %u keys for this missing inode\n", nr_keys);
-	else if (nr_keys > reconstruct_limit)
-		prt_printf(&buf, "found %u keys for this missing inode\n", nr_keys);
+		prt_printf(&buf, "found > %u keys for this inode\n", nr_keys);
+	else
+		prt_printf(&buf, "found %u keys for this inode\n", nr_keys);
+
+	if (c->sb.btrees_lost_data & BIT_ULL(BTREE_ID_inodes))
+		prt_str(&buf, "data was lost in inodes btree\n");
 
 	if (!have_inode) {
-		if (fsck_err_on(!have_inode,
-				trans, key_in_missing_inode,
-				"key in missing inode%s", buf.buf)) {
-			/*
-			 * Maybe a deletion that raced with data move, or something
-			 * weird like that? But if we know the inode was deleted, or
-			 * it's just a few keys, we can safely delete them.
-			 *
-			 * If it's many keys, we should probably recreate the inode
-			 */
-			if (have_old_inode || nr_keys <= 2)
-				goto delete;
-			else
-				goto reconstruct;
+		bool inode_looks_deleted =
+			good_ancestor &&
+			nr_keys < 3 &&
+			!(c->sb.btrees_lost_data & BIT_ULL(BTREE_ID_inodes));
+		if (inode_looks_deleted)
+			prt_str(&buf, "inode was deleted, will delete key\n");
+
+		if (ret_fsck_err(trans, key_in_missing_inode, "%s", buf.buf)) {
+			if (inode_looks_deleted)
+				return bch2_btree_delete_at(trans, iter, BTREE_UPDATE_internal_snapshot_node);
+
+			if (!good_ancestor) {
+				try(reconstruct_inode(trans, iter->btree_id, k.k->p.snapshot, k.k->p.inode));
+				try(bch2_trans_commit(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc));
+
+				inode->last_pos.inode--;
+				return bch_err_throw(c, transaction_restart_commit);
+			} else {
+				u32 snapshot = i->inode.bi_snapshot;
+				i->inode = good_ancestor->inode;
+				i->inode.bi_snapshot = snapshot;
+				try(bch2_fsck_write_inode(trans, &i->inode));
+				try(bch2_trans_commit_lazy(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc));
+			}
 		}
 	} else {
-		/*
-		 * not autofix, this one would be a giant wtf - bit error in the
-		 * inode corrupting i_mode?
-		 *
-		 * may want to try repairing inode instead of deleting
-		 */
-		if (fsck_err_on(!btree_matches_i_mode(iter->btree_id, i->inode.bi_mode),
-				trans, key_in_wrong_inode_type,
-				"key for wrong inode mode %o%s",
-				i->inode.bi_mode, buf.buf))
-			goto delete;
-	}
-out:
-err:
-fsck_err:
-	bch_err_fn(c, ret);
-	return ret;
-delete:
-	/*
-	 * XXX: print out more info
-	 * count up extents for this inode, check if we have different inode in
-	 * an older snapshot version, perhaps decide if we want to reconstitute
-	 */
-	ret = bch2_btree_delete_at(trans, iter, BTREE_UPDATE_internal_snapshot_node);
-	goto out;
-reconstruct:
-	ret =   reconstruct_inode(trans, iter->btree_id, k.k->p.snapshot, k.k->p.inode) ?:
-		bch2_trans_commit(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc);
-	if (ret)
-		goto err;
+		if (ret_fsck_err(trans, key_in_wrong_inode_type, "%s", buf.buf)) {
+			int nr_extents = iter->btree_id == BTREE_ID_extents
+				? nr_keys : count_inode_keys(trans, inode_pos, BTREE_ID_extents, NULL);
+			if (nr_extents < 0)
+				return nr_extents;
 
-	inode->last_pos.inode--;
-	ret = bch_err_throw(c, transaction_restart_nested);
-	goto out;
+			int nr_dirents = iter->btree_id == BTREE_ID_dirents
+				? nr_keys : count_inode_keys(trans, inode_pos, BTREE_ID_dirents, NULL);
+			if (nr_dirents < 0)
+				return nr_dirents;
+
+			if (nr_extents && nr_dirents) {
+				bch_err(c, "have both extents and dirents for inode with bad mode, cannot repair");
+				return bch_err_throw(c, shutdown_with_errors_unfixed);
+			}
+
+			i->inode.bi_mode &= ~S_IFMT;
+
+			if (nr_dirents)
+				i->inode.bi_mode |= S_IFDIR;
+			else
+				i->inode.bi_mode |= S_IFREG;
+
+			try(bch2_fsck_write_inode(trans, &i->inode));
+			try(bch2_trans_commit_lazy(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc));
+		}
+	}
+
+	return 0;
 }
 
 static int maybe_reconstruct_inum_btree(struct btree_trans *trans,
@@ -1602,8 +1615,15 @@ static int check_dirent(struct btree_trans *trans, struct btree_iter *iter,
 
 	hash_info->cf_encoding = bch2_inode_casefold(c, &i->inode) ? c->cf_encoding : NULL;
 
+	bool invalidated_inodes = false;
 	ret = bch2_str_hash_check_key(trans, s, &bch2_dirent_hash_desc, hash_info,
-				      iter, k, need_second_pass);
+				      k, need_second_pass, &invalidated_inodes);
+	if (invalidated_inodes) {
+		dir->last_pos.inode = 0;
+		dir->inodes.nr = 0;
+		return bch_err_throw(c, transaction_restart_nested);
+	}
+
 	if (ret < 0)
 		return ret;
 	if (ret)
@@ -1735,8 +1755,15 @@ static int check_xattr(struct btree_trans *trans, struct btree_iter *iter,
 	inode->first_this_inode = false;
 
 	bool need_second_pass = false;
-	return bch2_str_hash_check_key(trans, NULL, &bch2_xattr_hash_desc, hash_info,
-				      iter, k, &need_second_pass);
+	bool invalidated_inodes = false;
+	ret = bch2_str_hash_check_key(trans, NULL, &bch2_xattr_hash_desc, hash_info,
+				      k, &need_second_pass, &invalidated_inodes);
+	if (invalidated_inodes) {
+		inode->last_pos.inode--;
+		return bch_err_throw(c, transaction_restart_nested);
+	}
+
+	return min(ret, 0);
 }
 
 /*
