@@ -1233,7 +1233,11 @@ static CLOSURE_CALLBACK(bch2_journal_read_device)
 	ja->discard_idx = ja->dirty_idx_ondisk =
 		ja->dirty_idx = (ja->cur_idx + 1) % ja->nr;
 out:
-	bch_verbose(c, "journal read done on device %s, ret %i", ca->name, ret);
+	if (!ret)
+		bch_verbose_dev(ca, "journal read done");
+	else
+		bch_err_dev(ca, "journal read error %s", bch2_err_str(ret));
+
 	kvfree(buf.data);
 	enumerated_ref_put(&ca->io_ref[READ], BCH_DEV_READ_REF_journal_read);
 	closure_return(cl);
@@ -1247,33 +1251,30 @@ err:
 noinline_for_stack
 static void bch2_journal_print_checksum_error(struct bch_fs *c, struct journal_replay *j)
 {
-	CLASS(printbuf, buf)();
-	bch2_log_msg_start(c, &buf);
+	CLASS(bch_log_msg, msg)(c);
 
 	enum bch_csum_type csum_type = JSET_CSUM_TYPE(&j->j);
 	bool have_good = false;
 
-	prt_printf(&buf, "invalid journal checksum(s) at seq %llu ", le64_to_cpu(j->j.seq));
-	bch2_journal_datetime_to_text(&buf, &j->j);
-	prt_newline(&buf);
+	prt_printf(&msg.m, "invalid journal checksum(s) at seq %llu ", le64_to_cpu(j->j.seq));
+	bch2_journal_datetime_to_text(&msg.m, &j->j);
+	prt_newline(&msg.m);
 
 	darray_for_each(j->ptrs, ptr)
 		if (!ptr->csum_good) {
-			bch2_journal_ptr_to_text(&buf, c, ptr);
-			prt_char(&buf, ' ');
-			bch2_csum_to_text(&buf, csum_type, ptr->csum);
-			prt_newline(&buf);
+			bch2_journal_ptr_to_text(&msg.m, c, ptr);
+			prt_char(&msg.m, ' ');
+			bch2_csum_to_text(&msg.m, csum_type, ptr->csum);
+			prt_newline(&msg.m);
 		} else {
 			have_good = true;
 		}
 
-	prt_printf(&buf, "should be ");
-	bch2_csum_to_text(&buf, csum_type, j->j.csum);
+	prt_printf(&msg.m, "should be ");
+	bch2_csum_to_text(&msg.m, csum_type, j->j.csum);
 
 	if (have_good)
-		prt_printf(&buf, "\n(had good copy on another device)");
-
-	bch2_print_str(c, KERN_ERR, buf.buf);
+		prt_printf(&msg.m, "\n(had good copy on another device)");
 }
 
 struct u64_range bch2_journal_entry_missing_range(struct bch_fs *c, u64 start, u64 end)
@@ -1351,7 +1352,7 @@ int bch2_journal_read(struct bch_fs *c, struct journal_start_info *info)
 	struct journal_list jlist;
 	struct journal_replay *i, **_i;
 	struct genradix_iter radix_iter;
-	bool degraded = false, last_write_torn = false;
+	bool last_write_torn = false;
 	u64 seq;
 	int ret = 0;
 
@@ -1377,11 +1378,14 @@ int bch2_journal_read(struct bch_fs *c, struct journal_start_info *info)
 				     system_unbound_wq,
 				     &jlist.cl);
 		else
-			degraded = true;
+			set_bit(JOURNAL_degraded, &c->journal.flags);
 	}
 
-	while (closure_sync_timeout(&jlist.cl, sysctl_hung_task_timeout_secs * HZ / 2))
-		;
+	if (!sysctl_hung_task_timeout_secs)
+		closure_sync(&jlist.cl);
+	else
+		while (closure_sync_timeout(&jlist.cl, sysctl_hung_task_timeout_secs * HZ / 2))
+			;
 
 	if (jlist.ret)
 		return jlist.ret;
@@ -1515,17 +1519,6 @@ int bch2_journal_read(struct bch_fs *c, struct journal_start_info *info)
 			replicas_entry_add_dev(&replicas.e, ptr->dev);
 
 		bch2_replicas_entry_sort(&replicas.e);
-
-		CLASS(printbuf, buf)();
-		bch2_replicas_entry_to_text(&buf, &replicas.e);
-
-		if (!degraded &&
-		    !bch2_replicas_marked(c, &replicas.e) &&
-		    (le64_to_cpu(i->j.seq) == info->seq_read_start ||
-		     fsck_err(c, journal_entry_replicas_not_marked,
-			      "superblock not marked as containing replicas for journal entry %llu\n%s",
-			      le64_to_cpu(i->j.seq), buf.buf)))
-			try(bch2_mark_replicas(c, &replicas.e));
 	}
 fsck_err:
 	return ret;
